import dataclasses
import requests


@dataclasses.dataclass
class Flair:
    label: str
    name: str
    hidden: bool
    priority: int
    color: str
    rainbowColor: bool
    image: list[dict]

    def __repr__(self):
        return f"Flair<{self.label}>"


def flair_converter(endpoint: str) -> dict:
    """Returns a dict to convert flair names (e.g. flair17) to a Flair object."""
    r = requests.get(endpoint)
    converter = {item["name"]: Flair(**item) for item in r.json()}
    return converter



from .bot import DGGBot
from .chat import DGGChat
from .event import EventType
from .live import DGGLive, StreamInfo, YoutubeVideo, YoutubeVod
from .message import Message, PrivateMessage, MuteMessage



from datetime import datetime, timezone
import json
import requests
from typing import Union

from ._logging import _logger
from .event import EventType
from .flairs import Flair, flair_converter
from .funcs import threaded
from .message import (
    Message,
    MuteMessage,
    PinnedMessage,
    PollMessage,
    PrivateMessage,
    VoteMessage,
)
from .user import User
from .wsbase import WSBase
from .errors import (
    AccountTooYoung,
    Banned,
    DuplicateMessage,
    InvalidMessage,
    NeedLogin,
    NoPermission,
    NotFound,
    ProtocolError,
    SubMode,
    Throttled,
    TooManyConnections,
)


class DGGChat(WSBase):
    _CONFIG = {
        "wss": "wss://chat.destiny.gg/ws",
        "wss-origin": "https://www.destiny.gg",
        "baseurl": "https://www.destiny.gg",
        "endpoints": {"user": "/api/chat/me", "userinfo": "/api/userinfo"},
        "flairs": "https://cdn.destiny.gg/flairs/flairs.json",
    }

    def __init__(
        self,
        auth_token=None,
        wss: str = None,
        *,
        sid: str = None,
        rememberme: str = None,
        config: Union[str, dict[str, dict]] = None,
        **kwargs,
    ):
        self.username = None
        cookie = (
            f"authtoken={auth_token}"
            if auth_token
            else (
                f"sid={sid}" + (f";rememberme={rememberme}" if rememberme else "")
                if sid
                else None
            )
        )
        super().__init__(wss, cookie, config=config)
        if auth_token:
            self.username = self._get_username_from_token(auth_token)
        elif sid:
            self.username = self._get_username_from_sid(cookie)
        self._flairs = flair_converter(self.config["flairs"])
        self.authenticated = False
        self._users = {}

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(username='{self.username}')"
            if self.username
            else f"{self.__class__.__name__}()"
        )

    def _dggtime_to_dt(self, timestamp: str) -> datetime:
        timestamp = timestamp.replace("+0000", "Z")
        return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ")

    def _dggepoch_to_dt(self, epoch: int) -> datetime:
        return datetime.fromtimestamp(
            epoch // 1000, tz=timezone.utc
        )  # dgg sends milliseconds

    def _convert_flairs(self, data: dict) -> Union[list[Flair], None]:
        if "features" in data:
            return [self._flairs[flair] for flair in data["features"]]

    def _get_username_from_token(self, auth_token: str) -> Union[str, None]:
        r = requests.get(
            f"{self.config['baseurl']}{self.config['endpoints']['userinfo']}?token={auth_token}"
        )
        resp = r.json()
        if "username" in resp:
            return resp["username"].lower()
        else:
            _logger.error(
                f"Could not get username from token: {resp.get('message', 'Unknown reason.')}"
            )

    def _get_username_from_sid(self, cookie: str) -> Union[str, None]:
        headers = {"cookie": cookie}
        r = requests.get(
            self.config["baseurl"] + self.config["endpoints"]["user"],
            headers=headers,
        )
        resp = r.json()
        if r.status_code == 200:
            resp = r.json()
            return resp["nick"].lower()
        else:
            _logger.error(
                f"Authentication failed. Error {resp['code']}: {resp['error']}"
            )

    @property
    def users(self) -> dict:
        return self._users.copy()

    def get_user(self, username: str) -> Union[User, None]:
        return self._users.get(username.lower())

    def get_flair(self, name: str) -> Flair:
        return self._flairs[name]

    _err_dict = {
        "banned": Banned,
        "duplicate": DuplicateMessage,
        "invalidmsg": InvalidMessage,
        "needlogin": NeedLogin,
        "nopermission": NoPermission,
        "notfound": NotFound,
        "privmsgaccounttooyoung": AccountTooYoung,
        "protocolerror": ProtocolError,
        "submode": SubMode,
        "throttled": Throttled,
        "toomanyconnections": TooManyConnections,
    }

    def _on_message(self, ws, message: str):
        event_type, data = message.split(maxsplit=1)
        data = json.loads(data)
        if event_type == EventType.MUTE:
            msg = MuteMessage(
                self,
                event_type,
                data["nick"],
                self._dggtime_to_dt(data["createdDate"]),
                self._convert_flairs(data),
                self._dggepoch_to_dt(data["timestamp"]),
                data["data"],
                data["duration"],
            )
        elif event_type == EventType.PRIVMSG:
            msg = PrivateMessage(
                self,
                event_type,
                data["nick"],
                timestamp=self._dggepoch_to_dt(data["timestamp"]),
                data=data["data"],
                message_id=data["messageid"],
            )
        elif event_type == EventType.PRIVMSGSENT:
            return
        elif event_type == EventType.ERROR:
            if (desc := data["description"]) in self._err_dict:
                raise self._err_dict[desc]
            else:
                _logger.error(event_type, data)
                raise Exception(desc)
        elif event_type == EventType.NAMES:
            _logger.debug(f"{event_type} {data}")
            self.on_names(data["connectioncount"], data["users"])
            return
        elif event_type == EventType.PIN:
            msg = PinnedMessage(
                self,
                event_type,
                data["nick"],
                self._dggtime_to_dt(data["createdDate"]),
                self._convert_flairs(data),
                self._dggepoch_to_dt(data["timestamp"]),
                data.get("data"),
                data["uuid"],
            )
        elif event_type in (EventType.POLLSTART, EventType.POLLSTOP):
            msg = PollMessage(
                self,
                event_type,
                data["nick"],
                canvote=data["canvote"],
                myvote=data["myvote"],
                weighted=data["weighted"],
                start=self._dggtime_to_dt(data["start"]),
                now=self._dggtime_to_dt(data["now"]),
                # Time comes in milliseconds
                time=data["time"] // 1000,
                question=data["question"],
                options=data["options"],
                totals=data["totals"],
                totalvotes=data["totalvotes"],
            )
        elif event_type == EventType.VOTECAST:
            msg = VoteMessage(
                self,
                event_type,
                data["nick"],
                vote=data["vote"],
            )
        else:
            msg = Message(
                self,
                event_type,
                data.get("nick"),
                self._dggtime_to_dt(data["createdDate"])
                if "createdDate" in data
                else None,
                self._convert_flairs(data),
                self._dggepoch_to_dt(data["timestamp"]),
                data.get("data"),
            )

        if event_type == EventType.QUIT:
            self._users.pop(msg.nick_lower, None)
        elif event_type == EventType.JOIN:
            self._users[msg.nick_lower] = User(msg.nick, msg.createdDate, msg.features)

        self.on_event(event_type.lower(), msg)
        if event_type in (EventType.MESSAGE, EventType.PRIVMSG) and self.is_mentioned(
            msg
        ):
            self.on_event("mention", msg)
        self._post_message(msg)

    def _post_message(self, msg):
        """Do stuff after _on_message"""
        pass

    def _on_open(self, ws):
        _logger.info(
            f"Connecting "
            + (f"as {self.username} " if self.username else "")
            + f"to {self.wss}."
        )
        self._connected = True

    def mention(self):
        """Decorator to run function on mentions. Shortcut for event('on_mention')."""
        return self.event("on_mention")

    def is_mentioned(self, msg: Union[Message, PrivateMessage]) -> bool:
        return (
            False if self.username is None else (self.username in msg.data.casefold())
        )

    @threaded
    def on_names(self, connection_count: int, users: list):
        """Do stuff when the NAMES message is received upon connecting to chat."""
        self._users = {
            user["nick"].lower(): User(
                user["nick"],
                self._dggtime_to_dt(user["createdDate"]),
                user["features"],
            )
            for user in users
        }
        self.on_event("names", connection_count, users)

    @threaded
    def send(self, msg: str):
        """Send a message to chat."""
        payload = {"data": msg}
        self.ws.send(f"MSG {json.dumps(payload)}")

    @threaded
    def send_privmsg(self, nick: str, msg: str):
        """Send private message to someone."""
        payload = {"nick": nick, "data": msg}
        self.ws.send(f"PRIVMSG {json.dumps(payload)}")

    @threaded
    def cast_vote(self, vote: int):
        """Participate in a chat poll."""
        payload = {"vote": str(vote)}
        self.ws.send(f"CASTVOTE {json.dumps(payload)}")



import inspect
import itertools
from typing import Callable, Union
from .message import Message, PrivateMessage
from .chat import DGGChat, EventType


class DGGBot(DGGChat):
    def __init__(
        self,
        auth_token: str = None,
        owner: str = None,
        prefix="!",
        wss: str = None,
        *,
        sid: str = None,
        rememberme: str = None,
        **kwargs,
    ):
        super().__init__(
            auth_token=auth_token, wss=wss, sid=sid, rememberme=rememberme, **kwargs
        )
        self._owner = owner.lower() if owner else None
        self.prefix = prefix
        self._commands = {}

    def command(
        self,
        aliases: Union[list[str], tuple[str]] = tuple(),
        *args,
        cooldown: Union[int, float] = None,
        whisper_only: bool = False,
        **kwargs,
    ):
        """
        Decorator to add commands to bot.
        :param aliases: aliases to call the function besides the function name.
        :param cooldown: Cooldown (in seconds) between command uses.
        :param whisper_only: Command will only run when through whispers/private messages.
        :return:
        """

        def decorator(func: Callable):
            func._args = inspect.getfullargspec(func)
            if cooldown is not None:
                func._cooldown = cooldown
                func._last_called = 0
            if whisper_only:
                func = self.check(lambda msg: isinstance(msg, PrivateMessage))(func)
            for cmd_name in itertools.chain((func.__name__,), aliases):
                if cmd_name in self._commands:
                    raise Exception(f'Command name "{cmd_name}" already exists.')
                else:
                    self._commands[cmd_name] = func
            return func

        return decorator

    def is_command(self, msg: Message) -> bool:
        if isinstance(msg.data, str):
            return msg.data.startswith(self.prefix)

    def on_command(self, msg: Message):
        cmd = msg.data.split(" ")[0][1:]
        if cmd in self._commands:
            func: callable = self._commands[cmd]
            if hasattr(func, "_cooldown"):
                ts = int(msg.timestamp.timestamp())
                if (ts - func._last_called) < (func._cooldown):
                    return
                func._last_called = ts
            if hasattr(func, "_perms"):
                perms = func._perms
                if any(not perm(msg) for perm in perms):
                    return
            args = msg.data.split(" ", len(func._args.args) - (not func._args.varargs))[
                1:
            ]
            func(msg, *args)

    def check(self, *check_funcs: Callable):
        """
        Decorator to restrict command-usage by using a check function(s).
        :param check_funcs: functions that check for some statement and returns a bool.
        :return:
        """

        def decorator(func: Callable):
            if not hasattr(func, "_perms"):
                func._perms = []
            func._perms.extend(check_funcs)
            return func

        return decorator

    def is_owner(self):
        """Decorator to make command only usable by the bot owner."""
        return self.check(lambda msg: msg.nick.lower() == self._owner)

    def _post_message(self, msg: Message):
        if msg.type in (EventType.MESSAGE, EventType.PRIVMSG) and self.is_command(msg):
            self.on_command(msg)



class AccountTooYoung(Exception):
    pass


class Banned(Exception):
    pass


class DuplicateMessage(Exception):
    pass


class InvalidMessage(Exception):
    pass


class NeedLogin(Exception):
    pass


class NoPermission(Exception):
    pass


class NotFound(Exception):
    pass


class ProtocolError(Exception):
    pass


class SubMode(Exception):
    pass


class Throttled(Exception):
    pass


class TooManyConnections(Exception):
    pass



import json
from .message import StreamInfo, YoutubeVideo, YoutubeVod
from ..wsbase import WSBase
from .._logging import _logger


class DGGLive(WSBase):
    _CONFIG = {
        "wss": "wss://live.destiny.gg/ws",
        "wss-origin": "https://www.destiny.gg",
    }

    def __init__(self, auth_token=None, wss: str = None):
        super().__init__(wss, f"authtoken={auth_token}" if auth_token else None)
        self._live = False

    def __repr__(self):
        return f"{self.__class__.__name__}(live='{self.is_live}')"

    def is_live(self) -> bool:
        return self._live

    def set_live(self, state: bool):
        if not self.is_live() and state:
            self._live = state
            self.on_event("stream_start")
        elif self.is_live() and not state:
            self._live = state
            self.on_event("stream_end")

    def _on_message(self, ws, message: str):
        data = json.loads(message)
        event_type = data["type"]
        event_data = data
        if event_type == "dggApi:streamInfo":
            event_data = StreamInfo.from_json(data)
            self.set_live(event_data.live)
        elif event_type == "dggApi:youtubeVideos":
            event_data = YoutubeVideo.from_json(data)
        elif event_type == "dggApi:youtubeVods":
            event_data = YoutubeVod.from_json(data)
        self.on_event(event_type.split(":")[-1].lower(), event_data)



import logging

_logger = logging.getLogger("dgg-bot")
_logger.addHandler(logging.NullHandler())


def debug(msg, *args, **kwargs):
    _logger.debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    _logger.info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    _logger.warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    _logger.error(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    _logger.critical(msg, *args, **kwargs)



from functools import cache


class EventType:
    BAN = "BAN"
    BROADCAST = "BROADCAST"
    ERROR = "ERR"
    JOIN = "JOIN"
    MESSAGE = "MSG"
    MUTE = "MUTE"
    NAMES = "NAMES"
    PIN = "PIN"
    POLLSTART = "POLLSTART"
    POLLSTOP = "POLLSTOP"
    PRIVMSG = "PRIVMSG"
    PRIVMSGSENT = "PRIVMSGSENT"
    QUIT = "QUIT"
    REFRESH = "REFRESH"
    SUBONLY = "SUBONLY"
    UNBAN = "UNBAN"
    UNMUTE = "UNMUTE"
    VOTECAST = "VOTECAST"

    @classmethod
    @cache
    def types(cls) -> tuple[str]:
        return tuple(v for k, v in cls.__dict__.items() if k.isupper())



from abc import abstractmethod, ABC
import json
import time
from typing import Callable, Union
import websocket

from .funcs import threaded
from ._logging import _logger


class WSBase(ABC):
    _CONFIG = {
        "wss": "wss://chat.destiny.gg/ws",
        "wss-origin": "https://www.destiny.gg",
    }

    def __init__(
        self,
        wss: str = None,
        cookie: str = None,
        *,
        config: Union[str, dict[str, dict]] = None,
        **kwargs,
    ):
        if isinstance(config, str):
            with open(config) as f:
                self.config = json.load(f)
        elif config is not None:
            self.config = config
        else:
            self.config = self._CONFIG
        self.wss = wss or self.config["wss"]
        self.ws = websocket.WebSocketApp(
            self.wss,
            cookie=cookie,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )
        self._connected = False
        self._events = {}

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    # Event methods
    def event(self, event_name: str = None):
        """Decorator to run function when the specified event occurs."""

        def decorator(func: Callable):
            event = event_name or func.__name__
            if event not in self._events:
                self._events[event] = []
            self._events[event].append(func)
            return func

        return decorator

    @threaded
    def on_event(self, event: str, *args, **kwargs):
        for func in self._events.get(f"on_{event}", tuple()):
            func(*args, **kwargs)

    # Run methods
    def run(self, origin: str = None):
        self.ws.run_forever(origin=origin or self.config["wss-origin"])

    def run_forever(self, origin: str = None, sleep: int = 2):
        """Runs the client forever by automatically reconnecting the websocket."""
        while True:
            self.run(origin=origin or self.config["wss-origin"])
            time.sleep(sleep)

    # Websocket methods
    @abstractmethod
    def _on_message(self, ws, message: str):
        ...

    def _on_open(self, ws):
        _logger.info(f"Connecting to {self.wss}.")
        self._connected = True

    def _on_error(self, ws, error):
        _logger.error(error)

    def _on_close(self, ws, *_):
        _logger.debug(f"Connection closed.")
        self._connected = False



"""
Shared functions and whatnot for different classes.
"""
import threading


def threaded(func):
    """Decorator for making a threaded function."""

    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper



import dataclasses


@dataclasses.dataclass
class StreamInfo:
    live: bool
    game: str
    preview: str
    status_text: str
    started_at: str
    ended_at: str
    duration: int
    viewers: int
    id: str

    @classmethod
    def from_json(cls, data: dict) -> "StreamInfo":
        yt = data["data"]["streams"]["youtube"]
        del yt["platform"]
        del yt["type"]
        return cls(**yt)


@dataclasses.dataclass
class YoutubeVideo:
    id: str
    title: str
    mediumThumbnailUrl: str
    highThumbnailUrl: str
    url: str

    @classmethod
    def from_json(cls, data: dict) -> tuple["YoutubeVideo"]:
        return tuple(
            cls(
                vid["id"],
                vid["title"],
                vid["mediumThumbnailUrl"],
                vid["highThumbnailUrl"],
                vid["url"],
            )
            for vid in data["data"]
        )


@dataclasses.dataclass
class YoutubeVod:
    id: str
    title: str
    mediumThumbnailUrl: str
    highThumbnailUrl: str
    streamStartTime: str
    streamEndTime: str
    url: str

    @classmethod
    def from_json(cls, data: dict) -> tuple["YoutubeVod"]:
        return tuple(cls(**vod) for vod in data["data"])



from setuptools import setup, find_packages


with open("requirements.txt") as f:
    requirements = f.read().splitlines()


with open("README.rst") as f:
    readme = f.read()


packages = find_packages()

extra_require = {"dev": ["black", "flake8"]}


setup(
    name="dgg-bot",
    version="0.11.0",
    author="Fritz",
    description="A library for making a bot in Destiny.gg chat.",
    long_description=readme,
    long_description_content_type="text/x-rst",
    license="MIT",
    url="https://github.com/Fritz-02/dgg-bot/",
    project_urls={
        "Bug Tracker": "https://github.com/Fritz-02/dgg-bot/issues",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Communications :: Chat",
    ],
    install_requires=requirements,
    extra_require=extra_require,
    packages=packages,
    python_requires=">=3.9",
)



import dataclasses
from datetime import datetime
from ._logging import _logger
from .user import User


@dataclasses.dataclass
class _MessageBase:
    chat: "DGGChat"
    type: str
    nick: str = None

    @property
    def nick_lower(self) -> str:
        return self.nick.lower()

    def __post_init__(self):
        _logger.debug(self)


@dataclasses.dataclass
class Message(_MessageBase):
    createdDate: datetime = None
    features: list = None
    timestamp: datetime = None
    data: str = None

    @property
    def user(self) -> User:
        return User(self.nick, self.createdDate, self.features)

    def reply(self, content):
        self.chat.send(content)


@dataclasses.dataclass
class MuteMessage(Message):
    duration: int = None


@dataclasses.dataclass
class PinnedMessage(Message):
    uuid: str = None


@dataclasses.dataclass
class PrivateMessage(Message):
    message_id: str = None

    def reply(self, content):
        self.chat.send_privmsg(self.nick, content)


@dataclasses.dataclass
class PollMessage(_MessageBase):
    canvote: bool = None
    myvote: int = None
    weighted: bool = None
    start: datetime = None
    now: datetime = None
    # Poll time is in milliseconds
    time: int = None
    question: str = None
    options: list[str] = None
    totals: list[int] = None
    totalvotes: int = None

    def vote(self, option: int):
        self.chat.cast_vote(option)


@dataclasses.dataclass
class VoteMessage(_MessageBase):
    vote: str = None
